import java.util.*;

/* Models a linked list. */
public class MyLinkedList implements List {
  private Node myFront;
  private int mySize;

  /* Constructs a new empty list. */  
  public MyLinkedList() {
    myFront = null;
    mySize = 0;
  }

  /* Inserts the given element at the given index. */  
  public void add(int index, Object element) {
    checkIndex(index, size());
    
    if (index == 0) {
      myFront = new Node(element, myFront);
    } else {
      Node prev = getNodeAt(index - 1);
      prev.next = new Node(element, prev.next);
    }
    
    mySize++;
  }

  /* Appends the given element to the end of this list.  Returns true. */  
  public boolean add(Object element) {
    add(size(), element);
    return true;
  }
  
  /* Inserts the given element at the beginning of this list. */  
  public void addFirst(Object element) {
    add(0, element);
  }
  
  /* Appends the given element to the end of this list.  Identical to add(element). */  
  public void addLast(Object element) {
    add(element);
  }
  
  /* Removes all elements from this list. */  
  public void clear() {
    myFront = null;
  }
  
  /* Returns true if this list contains the given element. */  
  public boolean contains(Object element) {
    return indexOf(element) != -1;
  }
  
  /* Returns the element of this list at the given index.
     Throws IndexOutOfBoundsException if index is out of range.
  */
  public Object get(int index) {
    Node node = getNodeAt(index);
    return node.element;
  }
  
  /* Returns the first index where the given element occurs in this list, 
     or -1 if the element is not in this list.
  */
  // MODIFY THIS METHOD TO IMPROVE ITS PERFORMANCE ////////
  public int indexOf(Object element) {
    int index = -1;
    for (int i = 0; i < size(); i++) {
      Object currentElement = get(i);

      // if element is null, list's current element must be also.
      // if not, then they must be .equals to each other
      if (element == null) {
        if (currentElement == null) {
          index = i;
          break;
        }
      } else {
        if (currentElement != null && currentElement.equals(element)) {
          index = i;
          break;
        }
      }
    }
    
    return index;
  }
  
  /* Returns true if this list contains no elements. */
  public boolean isEmpty() {
    return myFront == null;
  }
  
  /* Returns the first index where the given element occurs in this list, 
     or -1 if the element is not in this list.
  */
  // MODIFY THIS METHOD TO IMPROVE ITS PERFORMANCE ////////
  public int lastIndexOf(Object element) {
    int index = -1;
    for (int i = 0; i < size(); i++) {
      Object currentElement = get(i);

      // if element is null, list's current element must be also.
      // if not, then they must be .equals to each other
      if (element == null) {
        if (currentElement == null)
          index = i;
      } else {
        if (currentElement != null && currentElement.equals(element))
          index = i;
      }
    }
    
    return index;
  }
  
  
  /* Removes the element of this list at the given index and returns it.
     Throws IndexOutOfBoundsException if index is out of range.
  */
  public Object remove(int index) {
    checkIndex(index, size() - 1);
    
    Object result = null;
    
    if (index == 0) {
      result = myFront.element;
      myFront = myFront.next;
    } else {
      Node prev = getNodeAt(index - 1);
      result = prev.next.element;
      prev.next = prev.next.next;
    }

    mySize--;
    return result;
  }
  
  /* Removes the given element from this list, if it is present in the list.
     Returns true if the element was in the list and was removed.
  */
  public boolean remove(Object element) {
    int index = indexOf(element);
    if (index != -1) {
      remove(index);
      return true;
    } else
      return false;
  }
  
  /*
  Removes from this list all of its elements that are contained in the specified collection.

  This implementation iterates over this list, checking each element returned by the iterator in turn to see if it's contained in the specified collection. If it's so contained, it's removed from this list with the iterator's remove method.
  
  Returns true if this list changed as a result of the call.
  */
  public boolean removeAll(Collection c) {
    boolean changed = false;

    ListIterator itr = listIterator();
    while (itr.hasNext()) {
      if (c.contains(itr.next())) {
        itr.remove();
        changed = true;
      }
    }
    
    return changed;
  }
  
  /* Sets the element of this list at the given index to have the given value.
     Throws IndexOutOfBoundsException if index is out of range.
  */
  public Object set(int index, Object value) {
    Node node = getNodeAt(index);
    node.element = value;
    return value;
  }
  
  /* Returns the number of elements in this list. */
  public int size() {
    return mySize;
  }
  
  /* You do not need to implement this method. */
  public List subList(int startIndex, int endIndex) {
    throw new UnsupportedOperationException();
  }

  /* Returns an array containing all of the elements in this list in the correct order. */
  public Object[] toArray() {
    Object[] array = new Object[size()];
    return toArray(array);
  }
  
  /* Returns the given array containing all of the elements in this list in the correct order. */
  // MODIFY THIS METHOD TO IMPROVE ITS PERFORMANCE ////////
  public Object[] toArray(Object[] array) {
    for (int i = 0; i < size(); i++)
      array[i] = get(i);
    
    return array;
  }

  /* Returns a String representation of this list. */  
  // MODIFY THIS METHOD TO IMPROVE ITS PERFORMANCE ////////
  public String toString() {
    String result = "[";

    if (!isEmpty()) {
      result += get(0);
      for (int i = 1; i < size(); i++) {
        result += ", " + get(i);
      }
    }

    result += "]";
    return result;
  }
  
  // METHODS FOR YOU TO IMPLEMENT /////////////////////////
  /*
  Inserts all of the elements in the specified collection into this list, starting at the specified position. Shifts the element currently at that position (if any) and any subsequent elements to the right (increases their indices). The new elements will appear in the list in the order that they are returned by the specified collection's iterator.
  
  Returns true if this list changed as a result of the call.  This will always be true if c has any elements inside it.
  */
  public boolean addAll(int index, Collection c) {
    throw new UnsupportedOperationException();
  }
  
  /*
  Appends all of the elements in the specified collection to the end of this list, in the order that they are returned by the specified collection's iterator.
  
  The behavior of this operation is undefined if the specified collection is modified while the operation is in progress. (This implies that the behavior of this call is undefined if the specified Collection is this list, and this list is nonempty.)
  
  Returns true if this list changed as a result of the call.  This will always be true if c has any elements inside it.
  */
  public boolean addAll(Collection c) {
    throw new UnsupportedOperationException();
  }
  
  /*
  Returns true if this collection contains all of the elements in the specified collection.
  
  This implementation iterates over the specified collection c, checking each element returned by the iterator in turn to see if it's contained in this list. If all elements are so contained true is returned, otherwise false. 
  */
  public boolean containsAll(Collection c) {
    throw new UnsupportedOperationException();
  }

  /*
  Returns an iterator over the elements in this list (in proper sequence).

  This implementation merely returns a list iterator over the list. 
  */
  public Iterator iterator() {
    throw new UnsupportedOperationException();
  }
  
  /*
  Returns an iterator of the elements in this list (in proper sequence).
  */
  public ListIterator listIterator() {
    throw new UnsupportedOperationException();
  }

  /*
  Returns a list iterator of the elements in this list (in proper sequence), starting at the specified position in this list. The specified index indicates the first element that would be returned by an initial call to the next method.
  */
  public ListIterator listIterator(int index) {
    throw new UnsupportedOperationException();
  }
  
  /*
  Retains only the elements in this list that are contained in the specified collection. In other words, removes from this list all of its elements that are not contained in the specified collection.

  This implementation iterates over this list, checking each element returned by the iterator in turn to see if it's contained in the specified collection. If it's not so contained, it's removed from this list with the iterator's remove method.
  
  Returns true if this list changed as a result of the call.
  */
  public boolean retainAll(Collection c) {
    throw new UnsupportedOperationException();
  }
  
  /////////////////////////////////////////////////////////
  
  
  /* Helper method: returns the node at the given index.  O(index). */
  private Node getNodeAt(int index) {
    checkIndex(index, size() - 1);

    Node curr = myFront;
    for (int i = 0; i < index; i++)
      curr = curr.next;
    
    return curr;
  }
  
  /* Throws an IndexOutOfBoundsException if 0 <= index <= max is not true.  O(1). */
  private void checkIndex(int index, int max) {
    if (index < 0 || index > max)
      throw new IndexOutOfBoundsException();
  }
  

  /* Represents one node in the list. */
  private class Node {
    /* The data element */
    public Object element;
    
    /* Reference to the next node in the list (null if none) */
    public Node next;
    
    /* Constructs a new node to store the given element, with no next node. */
    public Node(Object element) {
      this(element, null);
    }
    
    /* Constructs a new node to store the given element and the given next node. */
    public Node(Object element, Node next) {
      this.element = element;
      this.next = next;
    }
  }
  
/* Recommended iterator template.

  private class MyListIterator implements ListIterator {

    public MyListIterator(int index) {

    }
    
    public void add(Object element) {

    }
    
    public boolean hasNext() {

    }
    
    public Object next() {

    }
    
    public int nextIndex() {

    }
    
    public void remove() {

    }
    
    public void set(Object element) {

    }

    public boolean hasPrevious() {
      throw new UnsupportedOperationException();
    }
    
    public Object previous() {
      throw new UnsupportedOperationException();
    }
    
    public int previousIndex() {
      throw new UnsupportedOperationException();
    }    
  }
*/
}